/* camdrv.c */
/* CAMAC device driver for Hoshin CCP-PCI (with PCI-2703A adapter) */
/* Created by Enomoto Sanshiro on 11 April 1999. */
/* Updated by Enomoto Sanshiro on 7 January 2006. */
/* Updated by Enomoto Sanshiro on 28 December 2007. */


#if defined(USE_MODVERSIONS) && USE_MODVERSIONS
#  define MODVERSIONS
#  include <config/modversions.h>
#endif


#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include "camdrv.h"


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Enomoto Sanshiro");
MODULE_DESCRIPTION("Camac Driver for Hoshin CCP-PCI (with PCI-2703A) Controller");

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 36))
#define USE_OLD_IOCTL 1
#endif


#define PCI2703_VENDOR_ID 0x1147
#define PCI2703_DEVICE_ID 2703

#ifndef CAMDRV_MAJOR
#  define CAMDRV_MAJOR 0
#endif
static int camdrv_major = CAMDRV_MAJOR;
module_param(camdrv_major, int, S_IRUGO);

static char* camdrv_name = "camdrv";
static unsigned long ioport;
static struct pci_dev* camdrv_pci_dev = 0;

static int crate_number = 0;
static int open_count;

static int camdrv_open(struct inode* inode, struct file* filep);
static int camdrv_release(struct inode* inode, struct file* filep);
static int camdrv_pci_probe(struct pci_dev* dev, const struct pci_device_id* id);
static void camdrv_pci_remove(struct pci_dev* dev);
#ifdef USE_OLD_IOCTL
static int camdrv_ioctl(struct inode *inode, struct file *filep, unsigned int cmd, unsigned long arg);
#else
static long camdrv_unlocked_ioctl(struct file *filep, unsigned int cmd, unsigned long arg);
#endif

static int initialize(void);
static int clear(void);
static int inhibit(void);
static int release_inhibit(void);
static int camac_action(unsigned naf, unsigned* data);
static int read_lam(unsigned* data);
static int enable_interrupt(void);
static int disable_interrupt(void);
static int wait_lam(unsigned time_out, unsigned* data);

static void ccp_write(int address, int data);
static int ccp_read(int address);


static const struct pci_device_id camdrv_device_id_table[] = {
    { PCI_DEVICE(PCI2703_VENDOR_ID, PCI2703_DEVICE_ID) },
    { }
};
static struct pci_driver camdrv_pci_driver = {
    .id_table = camdrv_device_id_table,
    .probe = camdrv_pci_probe,
    .remove = camdrv_pci_remove,
};
static struct file_operations camdrv_fops = {
    .open = camdrv_open,
    .release = camdrv_release,
#ifdef USE_OLD_IOCTL
    .ioctl = camdrv_ioctl,
#else
    .unlocked_ioctl = camdrv_unlocked_ioctl,
#endif
};


int __init init_module(void)
{
    int result;

    camdrv_pci_dev = NULL;
    camdrv_pci_driver.name = camdrv_name;
    result = pci_register_driver(&camdrv_pci_driver);
    if (result < 0) {
        printk(KERN_WARNING "%s: unable to register CCP-PCI (PCI-2703).\n", camdrv_name);
	return result;
    }
    if (camdrv_pci_dev == NULL) {
        printk(KERN_WARNING "%s: unable to find CCP-PCI (PCI-2703).\n", camdrv_name);
	pci_unregister_driver(&camdrv_pci_driver);
	return -ENODEV;
    }

    result = register_chrdev(camdrv_major, camdrv_name, &camdrv_fops);
    if (result < 0) {
        printk(KERN_WARNING "%s: unable get major %d\n", camdrv_name, camdrv_major);
	pci_unregister_driver(&camdrv_pci_driver);
        return result;
    }
    if (camdrv_major == 0) {
        camdrv_major = result;
    }

    printk(KERN_INFO "%s: at 0x%04lx (major = %d).\n", camdrv_name, ioport, camdrv_major);

    open_count = 0;

    return 0;
}

void __exit cleanup_module(void)
{
    unregister_chrdev(camdrv_major, camdrv_name);
    pci_unregister_driver(&camdrv_pci_driver);
    printk(KERN_INFO "%s: removed.\n", camdrv_name);
}

static int camdrv_open(struct inode* inode, struct file* filep)
{
    if (open_count) {
        return -EBUSY;
    }

    try_module_get(THIS_MODULE);
    open_count++;

    return 0;
}

static int camdrv_release(struct inode* inode, struct file* filep)
{
    module_put(THIS_MODULE);
    open_count--;

    return 0;
}

static int camdrv_pci_probe(struct pci_dev* pci_dev, const struct pci_device_id* id)
{
    int result;

    printk(
	KERN_INFO "%s: CCP-PCI (PCI-2703) is detected at ioport 0x%04lx.\n",
	camdrv_name, 
	(unsigned long) pci_dev->resource[0].start
    );

    result = pci_enable_device(pci_dev);
    if (result < 0) {
	printk(KERN_WARNING "%s: enabling device failed.\n", camdrv_name);
	return result;
    }

    camdrv_pci_dev = pci_dev;
    ioport = pci_dev->resource[0].start & PCI_BASE_ADDRESS_IO_MASK;

    return 0;
}

static void camdrv_pci_remove(struct pci_dev* pci_dev)
{
    if (pci_dev == camdrv_pci_dev) {
	printk(KERN_WARNING "%s: device removed.\n", camdrv_name);
    }
}

#ifdef USE_OLD_IOCTL
static int camdrv_ioctl(struct inode* inode, struct file *filep, unsigned int cmd, unsigned long arg)
{
    return camdrv_unlocked_ioctl(filep, cmd, arg);
}
#endif

static long camdrv_unlocked_ioctl(struct file *filep, unsigned int cmd, unsigned long arg)
{
    unsigned parameter = 0, data = 0;
    unsigned *user_parameter_ptr, *user_data_ptr;
    int result = -EINVAL;

    user_parameter_ptr = (unsigned *) arg;
    user_data_ptr = (unsigned *) arg + 1;

    if (_IOC_TYPE(cmd) != CAMDRV_IOC_MAGIC) {
        return -EINVAL;
    }

    if (_IOC_DIR(cmd) & (_IOC_READ | _IOC_WRITE)) {
	if (get_user(parameter, user_parameter_ptr) < 0) {
	    return -EFAULT;
	}
	if (get_user(data, user_data_ptr) < 0) {
	    return -EFAULT;
	}
    }

    switch (cmd) {
      case CAMDRV_IOC_INITIALIZE:
        result = initialize();
        break;
      case CAMDRV_IOC_CLEAR:
        result = clear();
        break;
      case CAMDRV_IOC_INHIBIT:
        result = inhibit();
        break;
      case CAMDRV_IOC_RELEASE_INHIBIT:
        result = release_inhibit();
        break;
      case CAMDRV_IOC_ENABLE_INTERRUPT:
        result = enable_interrupt();
        break;
      case CAMDRV_IOC_DISABLE_INTERRUPT:
        result = disable_interrupt();
        break;
      case CAMDRV_IOC_CAMAC_ACTION:
        result = camac_action(parameter, &data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_READ_LAM:
        result = read_lam(&data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_WAIT_LAM:
        result = wait_lam(parameter, &data);
        put_user(data, user_data_ptr);
        break;
      case CAMDRV_IOC_SET_CRATE:
        crate_number = parameter;
        result = 0;
        break;
      default:
	result = -EINVAL;
    }

    return result;
}



/* CAMAC service functions */

enum ccp_regs {
    regDATA_LOW = 0x00,
    regDATA_MIDDLE = 0x01,
    regDATA_HIGH = 0x02,
    regLAM = 0x03,
    regSTATION = 0x03,
    regADDRESS = 0x04,
    regSTATUS = 0x05,
    regFUNCTION = 0x05,
    regOTHER_STATUS = 0x06,
    regOTHER_FUNCTION = 0x06,
    regENABLE_DEMAND = 0x07
};

enum ccp_ctrlbits {
    ctrlCLEAR = 0x80,
    ctrlINITIALIZE = 0x40,
    ctrlINHIBIT = 0x01,
    ctrlENABLE_DEMAND = 0x01,
    ctrlNONE = 0x00
};

enum ccp_statbits {
    statQ = 0x01,
    statX = 0x02,
    statDEMMAND = 0x40,
    statREQUEST = 0x80,
    statDEMAND_ENABLED = 0x01
};


static void ccp_write(int address, int data)
{
    address |= crate_number << 4;

    /* put ADDRESS and DATA values */
    outb_p(address, ioport + 2);
    outb_p(data, ioport + 1);
    wmb();

    /* make #WRITE strobe */
    outb(0x02, ioport + 3);
    outb(0x00, ioport + 3);
    wmb();

    /* 50 nsec delay is required here. */
#ifdef ENABLE_DELAY
    ndelay(50);
#endif

    /* reset DATA register. necessary?? */
    outb(0, ioport + 1);
}

static int ccp_read(int address)
{
    int data;

    address |= crate_number << 4;

    /* put ADDRESS value */
    outb_p(address, ioport + 2);

    /* make #READ strobe */
    outb(0x01, ioport + 3);
    wmb();

    /* 50 nsec delay is required here. */
#ifdef ENABLE_DELAY
    ndelay(50);
#endif

    /* read data */
    data = inb(ioport + 1);
    rmb();
    
    /* clear #READ strobe */
    outb(0x00, ioport + 3);

    return data;
}



static int initialize(void)
{
    ccp_write(regOTHER_FUNCTION, ctrlNONE);
    ccp_write(regFUNCTION, ctrlINITIALIZE);
    ccp_write(regOTHER_FUNCTION, ctrlNONE);

    ccp_write(regENABLE_DEMAND, ctrlENABLE_DEMAND);

    return 0;
}

static int clear(void)
{
    ccp_write(regFUNCTION, ctrlCLEAR);
    return 0;
}

static int inhibit(void)
{
    ccp_write(regOTHER_FUNCTION, ctrlINHIBIT);
    return 0;
}

static int release_inhibit(void)
{
    ccp_write(regOTHER_FUNCTION, ctrlNONE);
    return 0;
}

static int enable_interrupt(void)
{
    /* this function is not supported by the hardware */
    return 0;
}

static int disable_interrupt(void)
{
    /* this function is not supported by the hardware */
    return 0;
}

static int camac_action(unsigned naf, unsigned* data)
{
    unsigned long start_jiffies;
    int status, nq, nx;

    unsigned short data_l = (*data >> 0) & 0x00ff;
    unsigned short data_m = (*data >> 8) & 0x00ff;
    unsigned short data_h = (*data >> 16) & 0x00ff;

    unsigned short n = (naf >> 9) & 0x1f;
    unsigned short a = (naf >> 5) & 0x0f;
    unsigned short f = (naf >> 0) & 0x1f;

    ccp_write(regDATA_LOW, data_l);
    ccp_write(regDATA_MIDDLE, data_m);
    ccp_write(regDATA_HIGH, data_h);

    ccp_write(regSTATION, n);
    ccp_write(regADDRESS, a);
    ccp_write(regFUNCTION, f);

    start_jiffies = jiffies;
    while ((status = ccp_read(regSTATUS)) & statREQUEST) {
        schedule();
	if (jiffies - start_jiffies > HZ) {
	    *data = 0;
	    return -ETIMEDOUT;
	}
    }

    data_l = ccp_read(regDATA_LOW);
    data_m = ccp_read(regDATA_MIDDLE);
    data_h = ccp_read(regDATA_HIGH);

    *data = (data_h << 16) | (data_m << 8) | data_l;
    nq = (status & statQ) ? 0x00 : 0x01;
    nx = (status & statX) ? 0x00 : 0x01;

    return ((nx << 1) | nq);
}

static int read_lam(unsigned* data)
{
    unsigned short encoded_lam;
    encoded_lam = ccp_read(regLAM);

    if (encoded_lam > 0) {
        *data = 0x0001 << (encoded_lam - 1);
    }
    else {
        *data = 0;
    }

    return *data;
}

static int wait_lam(unsigned timeout, unsigned* data)
{
    unsigned long timeout_jiffies;
    unsigned short encoded_lam;

    /* The hardware does not support "interrupt on LAM". */
    /* The following code is a "polling loop" to wait for any LAM bits. */
    timeout_jiffies = jiffies + timeout * HZ;
    while ((encoded_lam = ccp_read(regLAM)) == 0) {
	schedule();
	if (jiffies > timeout_jiffies) {
	    *data = 0;
	    return -ETIMEDOUT;
	}
    }
    *data = 0x0001 << (encoded_lam - 1);

    return *data;
}
